@@ -2,17 +2,16 @@
   ==============================================================================
 
    This file is part of the JUCE library.
-   Copyright (c) 2017 - ROLI Ltd.
+   Copyright (c) 2022 - Raw Material Software Limited
 
    JUCE is an open source library subject to commercial or open-source
    licensing.
 
-   By using JUCE, you agree to the terms of both the JUCE 5 End-User License
-   Agreement and JUCE 5 Privacy Policy (both updated and effective as of the
-   27th April 2017).
+   By using JUCE, you agree to the terms of both the JUCE 7 End-User License
+   Agreement and JUCE Privacy Policy.
 
-   End User License Agreement: www.juce.com/juce-5-licence
-   Privacy Policy: www.juce.com/juce-5-privacy-policy
+   End User License Agreement: www.juce.com/juce-7-licence
+   Privacy Policy: www.juce.com/juce-privacy-policy
 
    Or: You may also use this code under the terms of the GPL v3 (see
    www.gnu.org/licenses).
@@ -27,55 +26,59 @@
 namespace juce
 {
 
-class ListBox::RowComponent  : public Component,
-                               public TooltipClient
+template <typename RowComponentType>
+static AccessibilityActions getListRowAccessibilityActions (RowComponentType& rowComponent)
 {
+    auto onFocus = [&rowComponent]
+    {
+        rowComponent.getOwner().scrollToEnsureRowIsOnscreen (rowComponent.getRow());
+        rowComponent.getOwner().selectRow (rowComponent.getRow());
+    };
+
+    auto onPress = [&rowComponent, onFocus]
+    {
+        onFocus();
+        rowComponent.getOwner().keyPressed (KeyPress (KeyPress::returnKey));
+    };
+
+    auto onToggle = [&rowComponent]
+    {
+        rowComponent.getOwner().flipRowSelection (rowComponent.getRow());
+    };
+
+    return AccessibilityActions().addAction (AccessibilityActionType::focus,  std::move (onFocus))
+                                 .addAction (AccessibilityActionType::press,  std::move (onPress))
+                                 .addAction (AccessibilityActionType::toggle, std::move (onToggle));
+}
+
+void ListBox::checkModelPtrIsValid() const
+{
+   #if ! JUCE_DISABLE_ASSERTIONS
+    // If this is hit, the model was destroyed while the ListBox was still using it.
+    // You should ensure that the model remains alive for as long as the ListBox holds a pointer to it.
+    // If this assertion is hit in the destructor of a ListBox instance, do one of the following:
+    // - Adjust the order in which your destructors run, so that the ListBox destructor runs
+    //   before the destructor of your ListBoxModel, or
+    // - Call ListBox::setModel (nullptr) before destroying your ListBoxModel.
+    jassert ((model == nullptr) == (weakModelPtr.lock() == nullptr));
+   #endif
+}
+
+//==============================================================================
+/*  The ListBox and TableListBox rows both have similar mouse behaviours, which are implemented here. */
+template <typename Base>
+class ComponentWithListRowMouseBehaviours : public Component
+{
+    auto& getOwner() const { return asBase().getOwner(); }
+
 public:
-    RowComponent (ListBox& lb) : owner (lb) {}
-
-    void paint (Graphics& g) override
+    void updateRowAndSelection (const int newRow, const bool nowSelected)
     {
-        if (auto* m = owner.getModel())
-            m->paintListBoxItem (row, g, getWidth(), getHeight(), selected);
-    }
+        const auto rowChanged       = std::exchange (row,      newRow)      != newRow;
+        const auto selectionChanged = std::exchange (selected, nowSelected) != nowSelected;
 
-    void update (const int newRow, const bool nowSelected)
-    {
-        if (row != newRow || selected != nowSelected)
-        {
+        if (rowChanged || selectionChanged)
             repaint();
-            row = newRow;
-            selected = nowSelected;
-        }
-
-        if (auto* m = owner.getModel())
-        {
-            setMouseCursor (m->getMouseCursorForRow (row));
-
-            customComponent.reset (m->refreshComponentForRow (newRow, nowSelected, customComponent.release()));
-
-            if (customComponent != nullptr)
-            {
-                addAndMakeVisible (customComponent.get());
-                customComponent->setBounds (getLocalBounds());
-            }
-        }
-    }
-
-    void performSelection (const MouseEvent& e, bool isMouseUp)
-    {
-        owner.selectRowsBasedOnModifierKeys (row, e.mods, isMouseUp);
-
-        if (auto* m = owner.getModel())
-            m->listBoxItemClicked (row, e);
-    }
-
-    bool isInDragToScrollViewport() const noexcept
-    {
-        if (auto* vp = owner.getViewport())
-            return vp->isScrollOnDragEnabled() && (vp->canScrollVertically() || vp->canScrollHorizontally());
-
-        return false;
     }
 
     void mouseDown (const MouseEvent& e) override
@@ -84,59 +87,121 @@ public:
         isDraggingToScroll = false;
         selectRowOnMouseUp = false;
 
-        if (isEnabled())
-        {
-            if (owner.selectOnMouseDown && ! (selected || isInDragToScrollViewport()))
-                performSelection (e, false);
-            else
-                selectRowOnMouseUp = true;
-        }
+        if (! asBase().isEnabled())
+            return;
+
+        const auto select = getOwner().getRowSelectedOnMouseDown()
+                            && ! selected
+                            && ! detail::ViewportHelpers::wouldScrollOnEvent (getOwner().getViewport(), e.source) ;
+        if (select)
+            asBase().performSelection (e, false);
+        else
+            selectRowOnMouseUp = true;
     }
 
     void mouseUp (const MouseEvent& e) override
     {
-        if (isEnabled() && selectRowOnMouseUp && ! (isDragging || isDraggingToScroll))
-            performSelection (e, true);
-    }
-
-    void mouseDoubleClick (const MouseEvent& e) override
-    {
-        if (isEnabled())
-            if (auto* m = owner.getModel())
-                m->listBoxItemDoubleClicked (row, e);
+        if (asBase().isEnabled() && selectRowOnMouseUp && ! (isDragging || isDraggingToScroll))
+            asBase().performSelection (e, true);
     }
 
     void mouseDrag (const MouseEvent& e) override
     {
-        if (auto* m = owner.getModel())
+        if (auto* m = getModel (getOwner()))
         {
-            if (isEnabled() && e.mouseWasDraggedSinceMouseDown() && ! isDragging)
+            if (asBase().isEnabled() && e.mouseWasDraggedSinceMouseDown() && ! isDragging)
             {
                 SparseSet<int> rowsToDrag;
 
-                if (owner.selectOnMouseDown || owner.isRowSelected (row))
-                    rowsToDrag = owner.getSelectedRows();
+                if (getOwner().getRowSelectedOnMouseDown() || getOwner().isRowSelected (row))
+                    rowsToDrag = getOwner().getSelectedRows();
                 else
                     rowsToDrag.addRange (Range<int>::withStartAndLength (row, 1));
 
-                if (rowsToDrag.size() > 0)
+                if (! rowsToDrag.isEmpty())
                 {
                     auto dragDescription = m->getDragSourceDescription (rowsToDrag);
 
                     if (! (dragDescription.isVoid() || (dragDescription.isString() && dragDescription.toString().isEmpty())))
                     {
                         isDragging = true;
-                        owner.startDragAndDrop (e, rowsToDrag, dragDescription, true);
+                        getOwner().startDragAndDrop (e, rowsToDrag, dragDescription, m->mayDragToExternalWindows());
                     }
                 }
             }
         }
 
         if (! isDraggingToScroll)
-            if (auto* vp = owner.getViewport())
+            if (auto* vp = getOwner().getViewport())
                 isDraggingToScroll = vp->isCurrentlyScrollingOnDrag();
     }
 
+    int getRow()            const { return row; }
+    bool isSelected()       const { return selected; }
+
+private:
+    const Base& asBase()    const { return *static_cast<const Base*> (this); }
+          Base& asBase()          { return *static_cast<      Base*> (this); }
+
+    static TableListBoxModel* getModel (TableListBox& x)     { return x.getTableListBoxModel(); }
+    static ListBoxModel*      getModel (ListBox& x)          { return x.getListBoxModel(); }
+
+    int row = -1;
+    bool selected = false, isDragging = false, isDraggingToScroll = false, selectRowOnMouseUp = false;
+};
+
+//==============================================================================
+class ListBox::RowComponent  : public TooltipClient,
+                               public ComponentWithListRowMouseBehaviours<RowComponent>
+{
+public:
+    explicit RowComponent (ListBox& lb) : owner (lb) {}
+
+    void paint (Graphics& g) override
+    {
+        if (auto* m = owner.getListBoxModel())
+            m->paintListBoxItem (getRow(), g, getWidth(), getHeight(), isSelected());
+    }
+
+    void update (const int newRow, const bool nowSelected)
+    {
+        updateRowAndSelection (newRow, nowSelected);
+
+        if (auto* m = owner.getListBoxModel())
+        {
+            setMouseCursor (m->getMouseCursorForRow (getRow()));
+
+            customComponent.reset (m->refreshComponentForRow (newRow, nowSelected, customComponent.release()));
+
+            if (customComponent != nullptr)
+            {
+                addAndMakeVisible (customComponent.get());
+                customComponent->setBounds (getLocalBounds());
+
+                setFocusContainerType (FocusContainerType::focusContainer);
+            }
+            else
+            {
+                setFocusContainerType (FocusContainerType::none);
+            }
+        }
+    }
+
+    void performSelection (const MouseEvent& e, bool isMouseUp)
+    {
+        owner.selectRowsBasedOnModifierKeys (getRow(), e.mods, isMouseUp);
+
+        if (auto* m = owner.getListBoxModel())
+            m->listBoxItemClicked (getRow(), e);
+    }
+
+    void mouseDoubleClick (const MouseEvent& e) override
+    {
+        if (isEnabled())
+            if (auto* m = owner.getListBoxModel())
+                m->listBoxItemDoubleClicked (getRow(), e);
+    }
+
     void resized() override
     {
         if (customComponent != nullptr)
@@ -145,63 +210,148 @@ public:
 
     String getTooltip() override
     {
-        if (auto* m = owner.getModel())
-            return m->getTooltipForRow (row);
+        if (auto* m = owner.getListBoxModel())
+            return m->getTooltipForRow (getRow());
 
         return {};
     }
 
+    std::unique_ptr<AccessibilityHandler> createAccessibilityHandler() override
+    {
+        return std::make_unique<RowAccessibilityHandler> (*this);
+    }
+
+    ListBox& getOwner() const { return owner; }
+
+    Component* getCustomComponent() const { return customComponent.get(); }
+
+private:
+    //==============================================================================
+    class RowAccessibilityHandler  : public AccessibilityHandler
+    {
+    public:
+        explicit RowAccessibilityHandler (RowComponent& rowComponentToWrap)
+            : AccessibilityHandler (rowComponentToWrap,
+                                    AccessibilityRole::listItem,
+                                    getListRowAccessibilityActions (rowComponentToWrap),
+                                    { std::make_unique<RowCellInterface> (*this) }),
+              rowComponent (rowComponentToWrap)
+        {
+        }
+
+        String getTitle() const override
+        {
+            if (auto* m = rowComponent.owner.getListBoxModel())
+                return m->getNameForRow (rowComponent.getRow());
+
+            return {};
+        }
+
+        String getHelp() const override  { return rowComponent.getTooltip(); }
+
+        AccessibleState getCurrentState() const override
+        {
+            if (auto* m = rowComponent.owner.getListBoxModel())
+                if (rowComponent.getRow() >= m->getNumRows())
+                    return AccessibleState().withIgnored();
+
+            auto state = AccessibilityHandler::getCurrentState().withAccessibleOffscreen();
+
+            if (rowComponent.owner.multipleSelection)
+                state = state.withMultiSelectable();
+            else
+                state = state.withSelectable();
+
+            if (rowComponent.isSelected())
+                state = state.withSelected();
+
+            return state;
+        }
+
+    private:
+        class RowCellInterface  : public AccessibilityCellInterface
+        {
+        public:
+            explicit RowCellInterface (RowAccessibilityHandler& h)  : handler (h)  {}
+
+            int getDisclosureLevel() const override  { return 0; }
+
+            const AccessibilityHandler* getTableHandler() const override
+            {
+                return handler.rowComponent.owner.getAccessibilityHandler();
+            }
+
+        private:
+            RowAccessibilityHandler& handler;
+        };
+
+        RowComponent& rowComponent;
+    };
+
+    //==============================================================================
     ListBox& owner;
     std::unique_ptr<Component> customComponent;
-    int row = -1;
-    bool selected = false, isDragging = false, isDraggingToScroll = false, selectRowOnMouseUp = false;
 
     JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (RowComponent)
 };
 
 
 //==============================================================================
-class ListBox::ListViewport  : public Viewport
+class ListBox::ListViewport  : public Viewport,
+                               private Timer
 {
 public:
     ListViewport (ListBox& lb)  : owner (lb)
     {
         setWantsKeyboardFocus (false);
 
-        auto content = new Component();
-        setViewedComponent (content);
+        struct IgnoredComponent : Component
+        {
+            std::unique_ptr<AccessibilityHandler> createAccessibilityHandler() override
+            {
+                return createIgnoredAccessibilityHandler (*this);
+            }
+        };
+
+        auto content = std::make_unique<IgnoredComponent>();
         content->setWantsKeyboardFocus (false);
+
+        setViewedComponent (content.release());
     }
 
-    RowComponent* getComponentForRow (const int row) const noexcept
+    int getIndexOfFirstVisibleRow() const { return jmax (0, firstIndex - 1); }
+
+    RowComponent* getComponentForRowIfOnscreen (int row) const noexcept
     {
-        return rows [row % jmax (1, rows.size())];
+        const auto startIndex = getIndexOfFirstVisibleRow();
+
+        return (startIndex <= row && row < startIndex + (int) rows.size())
+                 ? rows[(size_t) (row % jmax (1, (int) rows.size()))].get()
+                 : nullptr;
     }
 
-    RowComponent* getComponentForRowIfOnscreen (const int row) const noexcept
+    int getRowNumberOfComponent (const Component* const rowComponent) const noexcept
     {
-        return (row >= firstIndex && row < firstIndex + rows.size())
-                 ? getComponentForRow (row) : nullptr;
-    }
+        const auto iter = std::find_if (rows.begin(), rows.end(), [=] (auto& ptr) { return ptr.get() == rowComponent; });
 
-    int getRowNumberOfComponent (Component* const rowComponent) const noexcept
-    {
-        const int index = getViewedComponent()->getIndexOfChildComponent (rowComponent);
-        const int num = rows.size();
+        if (iter == rows.end())
+            return -1;
 
-        for (int i = num; --i >= 0;)
-            if (((firstIndex + i) % jmax (1, num)) == index)
-                return firstIndex + i;
+        const auto index = (int) std::distance (rows.begin(), iter);
+        const auto mod = jmax (1, (int) rows.size());
+        const auto startIndex = getIndexOfFirstVisibleRow();
 
-        return -1;
+        return index + mod * ((startIndex / mod) + (index < (startIndex % mod) ? 1 : 0));
     }
 
     void visibleAreaChanged (const Rectangle<int>&) override
     {
         updateVisibleArea (true);
 
-        if (auto* m = owner.getModel())
+        if (auto* m = owner.getListBoxModel())
             m->listWasScrolled();
+
+        startTimer (50);
     }
 
     void updateVisibleArea (const bool makeSureItUpdatesContent)
@@ -234,29 +384,33 @@ public:
             auto y = getViewPositionY();
             auto w = content.getWidth();
 
-            const int numNeeded = 2 + getMaximumVisibleHeight() / rowH;
-            rows.removeRange (numNeeded, rows.size());
+            const auto numNeeded = (size_t) (4 + getMaximumVisibleHeight() / rowH);
+            rows.resize (jmin (numNeeded, rows.size()));
 
             while (numNeeded > rows.size())
             {
-                auto newRow = new RowComponent (owner);
-                rows.add (newRow);
-                content.addAndMakeVisible (newRow);
+                rows.emplace_back (new RowComponent (owner));
+                content.addAndMakeVisible (*rows.back());
             }
 
             firstIndex = y / rowH;
             firstWholeIndex = (y + rowH - 1) / rowH;
             lastWholeIndex = (y + getMaximumVisibleHeight() - 1) / rowH;
 
-            for (int i = 0; i < numNeeded; ++i)
-            {
-                const int row = i + firstIndex;
+            const auto startIndex = getIndexOfFirstVisibleRow();
+            const auto lastIndex = startIndex + (int) rows.size();
 
-                if (auto* rowComp = getComponentForRow (row))
+            for (auto row = startIndex; row < lastIndex; ++row)
+            {
+                if (auto* rowComp = getComponentForRowIfOnscreen (row))
                 {
                     rowComp->setBounds (0, row * rowH, w, rowH);
                     rowComp->update (row, owner.isRowSelected (row));
                 }
+                else
+                {
+                    jassertfalse;
+                }
             }
         }
 
@@ -336,8 +490,21 @@ public:
     }
 
 private:
+    std::unique_ptr<AccessibilityHandler> createAccessibilityHandler() override
+    {
+        return createIgnoredAccessibilityHandler (*this);
+    }
+
+    void timerCallback() override
+    {
+        stopTimer();
+
+        if (auto* handler = owner.getAccessibilityHandler())
+            handler->notifyAccessibilityEvent (AccessibilityEvent::structureChanged);
+    }
+
     ListBox& owner;
-    OwnedArray<RowComponent> rows;
+    std::vector<std::unique_ptr<RowComponent>> rows;
     int firstIndex = 0, firstWholeIndex = 0, lastWholeIndex = 0;
     bool hasUpdated = false;
 
@@ -375,13 +542,16 @@ struct ListBoxMouseMoveSelector  : public MouseListener
 
 //==============================================================================
 ListBox::ListBox (const String& name, ListBoxModel* const m)
-    : Component (name), model (m)
+    : Component (name)
 {
     viewport.reset (new ListViewport (*this));
     addAndMakeVisible (viewport.get());
 
-    ListBox::setWantsKeyboardFocus (true);
-    ListBox::colourChanged();
+    setWantsKeyboardFocus (true);
+    setFocusContainerType (FocusContainerType::focusContainer);
+    colourChanged();
+
+    assignModelPtr (m);
 }
 
 ListBox::~ListBox()
@@ -390,11 +560,20 @@ ListBox::~ListBox()
     viewport.reset();
 }
 
+void ListBox::assignModelPtr (ListBoxModel* const newModel)
+{
+    model = newModel;
+
+   #if ! JUCE_DISABLE_ASSERTIONS
+    weakModelPtr = model != nullptr ? model->sharedState : nullptr;
+   #endif
+}
+
 void ListBox::setModel (ListBoxModel* const newModel)
 {
     if (model != newModel)
     {
-        model = newModel;
+        assignModelPtr (newModel);
         repaint();
         updateContent();
     }
@@ -458,6 +637,7 @@ Viewport* ListBox::getViewport() const noexcept
 //==============================================================================
 void ListBox::updateContent()
 {
+    checkModelPtrIsValid();
     hasDoneInitialUpdate = true;
     totalItems = (model != nullptr) ? model->getNumRows() : 0;
 
@@ -473,8 +653,14 @@ void ListBox::updateContent()
     viewport->updateVisibleArea (isVisible());
     viewport->resized();
 
-    if (selectionChanged && model != nullptr)
-        model->selectedRowsChanged (lastRowSelected);
+    if (selectionChanged)
+    {
+        if (model != nullptr)
+            model->selectedRowsChanged (lastRowSelected);
+
+        if (auto* handler = getAccessibilityHandler())
+            handler->notifyAccessibilityEvent (AccessibilityEvent::rowSelectionChanged);
+    }
 }
 
 //==============================================================================
@@ -488,6 +674,8 @@ void ListBox::selectRowInternal (const int row,
                                  bool deselectOthersFirst,
                                  bool isMouseClick)
 {
+    checkModelPtrIsValid();
+
     if (! multipleSelection)
         deselectOthersFirst = true;
 
@@ -509,6 +697,9 @@ void ListBox::selectRowInternal (const int row,
 
             lastRowSelected = row;
             model->selectedRowsChanged (row);
+
+            if (auto* handler = getAccessibilityHandler())
+                handler->notifyAccessibilityEvent (AccessibilityEvent::rowSelectionChanged);
         }
         else
         {
@@ -520,6 +711,8 @@ void ListBox::selectRowInternal (const int row,
 
 void ListBox::deselectRow (const int row)
 {
+    checkModelPtrIsValid();
+
     if (selected.contains (row))
     {
         selected.removeRange ({ row, row + 1 });
@@ -529,12 +722,17 @@ void ListBox::deselectRow (const int row)
 
         viewport->updateContents();
         model->selectedRowsChanged (lastRowSelected);
+
+        if (auto* handler = getAccessibilityHandler())
+            handler->notifyAccessibilityEvent (AccessibilityEvent::rowSelectionChanged);
     }
 }
 
 void ListBox::setSelectedRows (const SparseSet<int>& setOfRowsToBeSelected,
                                const NotificationType sendNotificationEventToModel)
 {
+    checkModelPtrIsValid();
+
     selected = setOfRowsToBeSelected;
     selected.removeRange ({ totalItems, std::numeric_limits<int>::max() });
 
@@ -545,6 +743,9 @@ void ListBox::setSelectedRows (const SparseSet<int>& setOfRowsToBeSelected,
 
     if (model != nullptr && sendNotificationEventToModel == sendNotification)
         model->selectedRowsChanged (lastRowSelected);
+
+    if (auto* handler = getAccessibilityHandler())
+        handler->notifyAccessibilityEvent (AccessibilityEvent::rowSelectionChanged);
 }
 
 SparseSet<int> ListBox::getSelectedRows() const
@@ -579,6 +780,8 @@ void ListBox::flipRowSelection (const int row)
 
 void ListBox::deselectAllRows()
 {
+    checkModelPtrIsValid();
+
     if (! selected.isEmpty())
     {
         selected.clear();
@@ -588,6 +791,9 @@ void ListBox::deselectAllRows()
 
         if (model != nullptr)
             model->selectedRowsChanged (lastRowSelected);
+
+        if (auto* handler = getAccessibilityHandler())
+            handler->notifyAccessibilityEvent (AccessibilityEvent::rowSelectionChanged);
     }
 }
 
@@ -655,12 +861,12 @@ int ListBox::getInsertionIndexForPosition (const int x, const int y) const noexc
 Component* ListBox::getComponentForRowNumber (const int row) const noexcept
 {
     if (auto* listRowComp = viewport->getComponentForRowIfOnscreen (row))
-        return listRowComp->customComponent.get();
+        return listRowComp->getCustomComponent();
 
     return nullptr;
 }
 
-int ListBox::getRowNumberOfComponent (Component* const rowComponent) const noexcept
+int ListBox::getRowNumberOfComponent (const Component* const rowComponent) const noexcept
 {
     return viewport->getRowNumberOfComponent (rowComponent);
 }
@@ -705,6 +911,8 @@ void ListBox::scrollToEnsureRowIsOnscreen (const int row)
 //==============================================================================
 bool ListBox::keyPressed (const KeyPress& key)
 {
+    checkModelPtrIsValid();
+
     const int numVisibleRows = viewport->getHeight() / getRowHeight();
 
     const bool multiple = multipleSelection
@@ -723,7 +931,7 @@ bool ListBox::keyPressed (const KeyPress& key)
         if (multiple)
             selectRangeOfRows (lastRowSelected, lastRowSelected + 1);
         else
-            selectRow (jmin (totalItems - 1, jmax (0, lastRowSelected) + 1));
+            selectRow (jmin (totalItems - 1, jmax (0, lastRowSelected + 1)));
     }
     else if (key.isKeyCode (KeyPress::pageUpKey))
     {
@@ -792,13 +1000,13 @@ void ListBox::mouseWheelMove (const MouseEvent& e, const MouseWheelDetails& whee
 {
     bool eventWasUsed = false;
 
-    if (wheel.deltaX != 0.0f && getHorizontalScrollBar().isVisible())
+    if (! approximatelyEqual (wheel.deltaX, 0.0f) && getHorizontalScrollBar().isVisible())
     {
         eventWasUsed = true;
         getHorizontalScrollBar().mouseWheelMove (e, wheel);
     }
 
-    if (wheel.deltaY != 0.0f && getVerticalScrollBar().isVisible())
+    if (! approximatelyEqual (wheel.deltaY, 0.0f) && getVerticalScrollBar().isVisible())
     {
         eventWasUsed = true;
         getVerticalScrollBar().mouseWheelMove (e, wheel);
@@ -810,6 +1018,8 @@ void ListBox::mouseWheelMove (const MouseEvent& e, const MouseWheelDetails& whee
 
 void ListBox::mouseUp (const MouseEvent& e)
 {
+    checkModelPtrIsValid();
+
     if (e.mouseWasClicked() && model != nullptr)
         model->backgroundClicked (e);
 }
@@ -856,14 +1066,18 @@ void ListBox::setOutlineThickness (int newThickness)
     resized();
 }
 
-void ListBox::setHeaderComponent (Component* newHeaderComponent)
+void ListBox::setHeaderComponent (std::unique_ptr<Component> newHeaderComponent)
 {
-    if (headerComponent.get() != newHeaderComponent)
-    {
-        headerComponent.reset (newHeaderComponent);
-        addAndMakeVisible (newHeaderComponent);
-        ListBox::resized();
-    }
+    headerComponent = std::move (newHeaderComponent);
+    addAndMakeVisible (headerComponent.get());
+    ListBox::resized();
+    invalidateAccessibilityHandler();
+}
+
+bool ListBox::hasAccessibleHeaderComponent() const
+{
+    return headerComponent != nullptr
+            && headerComponent->getAccessibilityHandler() != nullptr;
 }
 
 void ListBox::repaintRow (const int rowNumber) noexcept
@@ -871,7 +1085,7 @@ void ListBox::repaintRow (const int rowNumber) noexcept
     repaint (getRowPosition (rowNumber, true));
 }
 
-Image ListBox::createSnapshotOfRows (const SparseSet<int>& rows, int& imageX, int& imageY)
+ScaledImage ListBox::createSnapshotOfRows (const SparseSet<int>& rows, int& imageX, int& imageY)
 {
     Rectangle<int> imageArea;
     auto firstRow = getRowContainingPosition (0, viewport->getY());
@@ -893,8 +1107,12 @@ Image ListBox::createSnapshotOfRows (const SparseSet<int>& rows, int& imageX, in
     imageX = imageArea.getX();
     imageY = imageArea.getY();
 
-    auto listScale = Component::getApproximateScaleFactorForComponent (this);
-    Image snapshot (Image::ARGB, roundToInt (imageArea.getWidth() * listScale), roundToInt (imageArea.getHeight() * listScale), true);
+    const auto additionalScale = 2.0f;
+    const auto listScale = Component::getApproximateScaleFactorForComponent (this) * additionalScale;
+    Image snapshot (Image::ARGB,
+                    roundToInt ((float) imageArea.getWidth() * listScale),
+                    roundToInt ((float) imageArea.getHeight() * listScale),
+                    true);
 
     for (int i = getNumRowsOnScreen() + 2; --i >= 0;)
     {
@@ -903,9 +1121,9 @@ Image ListBox::createSnapshotOfRows (const SparseSet<int>& rows, int& imageX, in
             if (auto* rowComp = viewport->getComponentForRowIfOnscreen (firstRow + i))
             {
                 Graphics g (snapshot);
-                g.setOrigin (getLocalPoint (rowComp, Point<int>()) - imageArea.getPosition());
+                g.setOrigin ((getLocalPoint (rowComp, Point<int>()) - imageArea.getPosition()) * additionalScale);
 
-                auto rowScale = Component::getApproximateScaleFactorForComponent (rowComp);
+                const auto rowScale = Component::getApproximateScaleFactorForComponent (rowComp) * additionalScale;
 
                 if (g.reduceClipRegion (rowComp->getLocalBounds() * rowScale))
                 {
@@ -918,7 +1136,7 @@ Image ListBox::createSnapshotOfRows (const SparseSet<int>& rows, int& imageX, in
         }
     }
 
-    return snapshot;
+    return { snapshot, additionalScale };
 }
 
 void ListBox::startDragAndDrop (const MouseEvent& e, const SparseSet<int>& rowsToDrag, const var& dragDescription, bool allowDraggingToOtherWindows)
@@ -939,14 +1157,89 @@ void ListBox::startDragAndDrop (const MouseEvent& e, const SparseSet<int>& rowsT
     }
 }
 
+std::unique_ptr<AccessibilityHandler> ListBox::createAccessibilityHandler()
+{
+    class TableInterface  : public AccessibilityTableInterface
+    {
+    public:
+        explicit TableInterface (ListBox& listBoxToWrap)
+            : listBox (listBoxToWrap)
+        {
+        }
+
+        int getNumRows() const override
+        {
+            listBox.checkModelPtrIsValid();
+
+            return listBox.model != nullptr ? listBox.model->getNumRows()
+                                            : 0;
+        }
+
+        int getNumColumns() const override
+        {
+            return 1;
+        }
+
+        const AccessibilityHandler* getHeaderHandler() const override
+        {
+            if (listBox.hasAccessibleHeaderComponent())
+                return listBox.headerComponent->getAccessibilityHandler();
+
+            return nullptr;
+        }
+
+        const AccessibilityHandler* getRowHandler (int row) const override
+        {
+            if (auto* rowComponent = listBox.viewport->getComponentForRowIfOnscreen (row))
+                return rowComponent->getAccessibilityHandler();
+
+            return nullptr;
+        }
+
+        const AccessibilityHandler* getCellHandler (int, int) const override
+        {
+            return nullptr;
+        }
+
+        Optional<Span> getRowSpan (const AccessibilityHandler& handler) const override
+        {
+            const auto rowNumber = listBox.getRowNumberOfComponent (&handler.getComponent());
+
+            return rowNumber != -1 ? makeOptional (Span { rowNumber, 1 })
+                                   : nullopt;
+        }
+
+        Optional<Span> getColumnSpan (const AccessibilityHandler&) const override
+        {
+            return Span { 0, 1 };
+        }
+
+        void showCell (const AccessibilityHandler& h) const override
+        {
+            if (const auto row = getRowSpan (h))
+                listBox.scrollToEnsureRowIsOnscreen (row->begin);
+        }
+
+    private:
+        ListBox& listBox;
+
+        JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (TableInterface)
+    };
+
+    return std::make_unique<AccessibilityHandler> (*this,
+                                                   AccessibilityRole::list,
+                                                   AccessibilityActions{},
+                                                   AccessibilityHandler::Interfaces { std::make_unique<TableInterface> (*this) });
+}
+
 //==============================================================================
-Component* ListBoxModel::refreshComponentForRow (int, bool, Component* existingComponentToUpdate)
+Component* ListBoxModel::refreshComponentForRow (int, bool, [[maybe_unused]] Component* existingComponentToUpdate)
 {
-    ignoreUnused (existingComponentToUpdate);
     jassert (existingComponentToUpdate == nullptr); // indicates a failure in the code that recycles the components
     return nullptr;
 }
 
+String ListBoxModel::getNameForRow (int rowNumber)                      { return "Row " + String (rowNumber + 1); }
 void ListBoxModel::listBoxItemClicked (int, const MouseEvent&) {}
 void ListBoxModel::listBoxItemDoubleClicked (int, const MouseEvent&) {}
 void ListBoxModel::backgroundClicked (const MouseEvent&) {}
